from collections import namedtuple

from ..analysis import collect_var_names_from_exprs
from ..ndtypes import TupleT 
from ..syntax import (Return, Map, OuterMap, Tuple, Var, Closure, Assign, 
                      Const, TypedFn, UntypedFn, Ravel, Shape, Strides, Transpose, 
                      Where, Reshape) 
from ..syntax.helpers import none   

from subst import subst_expr_list
from transform import Transform

class CombineNestedMaps(Transform):

  def translate_expr(self, expr, mapping, forbidden = set([])):
      c = expr.__class__ 
      if c is Var:
        if expr.name in forbidden:
          return None 
        return mapping[expr.name]
      elif c in (Const, TypedFn, UntypedFn):
        return expr 
      
      elif c is Tuple:
        return Tuple(elts = tuple(self.translate_expr(elt,mapping,forbidden) 
                                  for elt in expr.elts), type = expr.type)
      elif c is Closure:
        return Closure(fn = expr.fn, 
                       args = tuple(self.translate_expr(elt, mapping,forbidden) 
                                    for elt in expr.args), type = expr.type)
      elif c in (Ravel, Shape, Strides, Transpose, Where):
        return c(array = self.translate_expr(expr.array, mapping,forbidden), type = expr.type)
      elif c is Reshape:
        return Reshape(array = self.translate_expr(expr.array, mapping,forbidden), 
                       shape = self.translate_expr(expr.shape, mapping,forbidden), 
                       type = expr.type)
  
  def build_arg_mapping(self, fn, closure_elts, outer_args):
    
    combined_args = tuple(closure_elts) + tuple(outer_args)
    assert len(fn.arg_names) == len(combined_args), \
      "Mismatch between function with %d formals and given %d actual args" % \
      (len(fn.arg_names), len(combined_args))
    mapping = {}
    for name, expr in zip(fn.arg_names, combined_args):
      mapping[name] = expr  
  
    # don't allow transformations of array elements
    # since that's an interaction between levels of the nested
    # maps 
    forbidden = set(fn.arg_names[len(closure_elts):])
    for stmt in fn.body[:-1]:
      if stmt.__class__ is Assign and stmt.lhs.__class__ is Var:
        new_rhs = self.translate_expr(stmt.rhs, mapping,forbidden)
        if new_rhs is None:
          return None 
        else:
          mapping[stmt.lhs.name] = new_rhs
      else:
        return None 
    return mapping 
  
  def combine_maps(self, closure, outer_args, outer_axis, result_type):
    fn = self.get_fn(closure)
    closure_elts = self.closure_elts(closure)
    
    if len(fn.body) == 0: return None 
    stmt = fn.body[-1]
    if stmt.__class__ is not Return: return None 
    nested_expr = stmt.value 
    
    if nested_expr.__class__ not in (Map, OuterMap): return None
    
    arg_mapping = self.build_arg_mapping(fn, closure_elts, outer_args)
    if arg_mapping is None: return None 
    
    nested_outer_args = nested_expr.args
    n_nested_outer_args = len(nested_outer_args)
    nested_map_closure_elts = self.closure_elts(nested_expr.fn) 
    nested_fn = self.get_fn(nested_expr.fn)
    nested_axis = nested_expr.axis
    
    # either both Maps specify an axis of traversal or neither do  
    if self.is_none(outer_axis) != self.is_none(nested_axis): return None 
    
    
    n_array_args = len(outer_args)
    
    inner_array_names = fn.arg_names[-n_array_args:]
    
    if len(nested_map_closure_elts) < len(inner_array_names): return None 
    
    # if any of the last k closure arguments aren't array elements
    # then abandon ship 
    k = n_array_args
    for closure_expr in nested_map_closure_elts[-k:]:
      if closure_expr.__class__ is not Var or closure_expr.name not in inner_array_names:
        return None 
      
    # if the two Maps are both elementwise, then make the OuterMap 
    # also elementwise

    remapped_inner_outer_args = [self.translate_expr(e,arg_mapping)
                               for e in nested_outer_args]
    remapped_inner_closure_elts = [self.translate_expr(e, arg_mapping)
                                   for e in nested_map_closure_elts]
    
    new_closure_elts = remapped_inner_closure_elts[:-k]
    new_outer_args = remapped_inner_closure_elts[-k:] + remapped_inner_outer_args    
    

    if self.is_none(outer_axis):
      combined_axis = none
    else:
      # combine the axes since we're going to make a single OuterMap 
      if isinstance(outer_axis.type, TupleT):
        outer_axes = self.tuple_elts(outer_axis)
      else:
        outer_axes = (outer_axis,) * n_array_args
      if isinstance(nested_axis.type, TupleT):
        assert isinstance(nested_axis, Tuple),  \
          "Axis argument must be None, int, or literal tuple, not %s" % nested_axis 
        inner_axes = tuple(nested_axis.elts) 
      else:
        inner_axes = (nested_axis,) * n_nested_outer_args
       
      combined_axes = list(outer_axes + inner_axes)
      for i, array_arg in enumerate(new_outer_args):
        if array_arg.__class__ is Transpose:
          old_axis = combined_axes[i]
          if not self.is_none(old_axis):
            combined_axes[i] = Const(value = array_arg.type.rank - old_axis.value - 1, 
                                     type = old_axis.type)
            new_outer_args[i] = array_arg.array
      combined_axis = self.tuple(combined_axes)
    return  OuterMap(fn = self.closure(nested_fn, new_closure_elts),
                     args = tuple(new_outer_args),   
                     axis = combined_axis, 
                     type = result_type)
    
  def transform_Map(self, expr):
    # can't turn Map(-, x, y) into an OuterMap since (x,y) are at the same iteration level 
    if len(expr.args) != 1:
      return expr
    new_expr = self.combine_maps(expr.fn, expr.args, expr.axis, expr.type)
    if new_expr is None: return expr 
    else: return new_expr 
    
  def transform_OuterMap(self, expr):
    new_expr = self.combine_maps(expr.fn, expr.args, expr.axis, expr.type)
    if new_expr is None: return expr 
    else: return new_expr 
    
    


import numpy as np 
from parakeet import jit 
# Simple convolution of 3x3 patches from a given array x
# by a 3x3 array of filter weights
 
@jit
def conv_3x3_trim(x, weights):
  return np.array([[(x[i-1:i+2, j-1:j+2]*weights).sum() 
                    for j in xrange(1, x.shape[1] -2)]
                    for i in xrange(1, x.shape[0] -2)])
 

x = np.random.randn(50,50)
w = np.random.randn(3,3)

print "Input", x
print "Convolved output", conv_3x3_trim(x,w)






from ..ndtypes import (make_slice_type, make_array_type, ptr_type, 
                       ArrayT, TupleT, ScalarT, Type, PtrT, Int64, IntT, Float64)
from ..syntax import (Alloc, AllocArray, ArrayView, Const, Index, Slice, Struct, Var, Select, Expr)
from ..syntax.helpers import (const, zero_i64, wrap_if_constant, slice_none, unwrap_constant)
from arith_builder import ArithBuilder

class ArrayBuilder(ArithBuilder):
  """
  Builder for constructing arrays and getting their properties
  """
  
  def elt_type(self, x):
    if isinstance(x, Type):
      try:
        return x.elt_type
      except:
        return x
    elif self.is_array(x):
      return x.type.elt_type
    else:
      return x.type
  
  def alloc_array(self, elt_t, dims, name = "array", 
                   explicit_struct = False, 
                   array_view = False, 
                   order = "C"):
    """
    Given an element type and sequence of expressions denoting each dimension
    size, generate code to allocate an array and its shape/strides metadata. For
    now I'm assuming that all arrays are in row-major, eventually we should make
    the layout an option.
    """
    
    assert order == "C", "Only row-major layout supported so far, not %s" % order 

    if self.is_tuple(dims):
      shape = dims
      dims = self.tuple_elts(shape)
    else:
      if not isinstance(dims, (list, tuple)):
        dims = [dims]
      shape = self.tuple(dims, "shape", explicit_struct = explicit_struct)
    rank = len(dims)
    array_t = make_array_type(elt_t, rank)
    if explicit_struct or array_view:
      nelts = self.prod(dims, name = "nelts")
      ptr_t = ptr_type(elt_t)

      ptr_var = self.assign_name(Alloc(elt_t, nelts, type = ptr_t), "data_ptr")
      stride_elts = [const(1)]

      # assume row-major for now!

      for d in reversed(dims[1:]):
        next_stride = self.mul(stride_elts[0], d, "dim")
        stride_elts = [next_stride] + stride_elts
      strides = self.tuple(stride_elts, "strides", explicit_struct = explicit_struct)
      if explicit_struct:
        array = Struct([ptr_var, shape, strides, zero_i64, nelts], type = array_t)
      else:
        array = ArrayView(data = ptr_var, 
                          shape = shape, 
                          strides = strides,
                          offset = zero_i64,
                          size = nelts, 
                          type = array_t)
    else:
      array = AllocArray(shape, elt_type = elt_t, type = array_t)
    if name is None: 
      return array 
    return self.assign_name(array, name)
  
  def len(self, array):
    return self.shape(array, 0)
  
  def nelts(self, array, explicit_struct = False):
    shape_elts = self.tuple_elts(self.shape(array), explicit_struct = explicit_struct)
    return self.prod(shape_elts, name = "nelts") 

  def rank(self, value):
    if self.is_array(value):
      return value.type.rank
    else:
      return 0

  def shape(self, array, dim = None, explicit_struct = False, temp = True):
    if array.type.__class__ is ArrayT:
      shape = self.attr(array, "shape", temp = temp)
      if dim is None:
        return shape
      if isinstance(dim, Expr):
        dim = unwrap_constant(dim)
      assert isinstance(dim, (int, long)), "Expected array dimension to be an int, got %s" % dim 
      dim_value = self.tuple_proj(shape, dim, explicit_struct = explicit_struct)
      if temp:
        return self.assign_name(dim_value, "dim%d" % dim)
      else:
        return dim_value 
    else:
      return self.tuple([])

  def strides(self, array, dim = None, explicit_struct = False):
    assert array.type.__class__ is ArrayT
    strides = self.attr(array, "strides")
    if dim is None:
      return strides
    else:
      elt_value = self.tuple_proj(strides, dim, explicit_struct = explicit_struct)
      return self.assign_name(elt_value, "stride%d" % dim)

  def slice_value(self, start, stop, step):
    slice_t = make_slice_type(start.type, stop.type, step.type)
    return Slice(start, stop, step, type = slice_t)

  def build_slice_indices(self, rank, axis, idx):
    """
    Build index tuple to pull out the 'idx' element along the given axis
    """
    if rank == 1:
      assert axis == 0
      return idx
    
    indices = []
    for i in xrange(rank):
      if i == axis:
        indices.append(idx)
      else:
        s = self.slice_value(self.none, self.none, self.int(1))
        indices.append(s)
    return self.tuple(indices)

  def elts_in_slice(self, start, stop, step):
    start_minus_start = self.sub(stop, start, name = "stop_minus_start")
    nelts = self.div(self.cast(start_minus_start, Float64), step, name = "nelts")
    ceil = self.ceil(nelts)
    nelts = self.cast(ceil, Int64)
    return self.max(nelts, self.zero(nelts.type))

  def slice_along_axis(self, arr, axis, idx):
    """
    Pull out a slice if the array has the given axis, 
    otherwise just return the array 
    """
    r = self.rank(arr)
    
    if isinstance(axis, Expr):
      axis = unwrap_constant(axis)
    idx = wrap_if_constant(idx)
    
    if r == 1 and (axis is None or axis == 0):
      return self.index(arr, idx)
    elif axis is None:
      if isinstance(idx.type, ScalarT):
        idx = self.tuple((idx,) * r)
      return self.index(arr, idx)
    elif r > axis:
      index_tuple = self.build_slice_indices(r, axis, idx)
      return self.index(arr, index_tuple)
    else:
      return arr

  def output_slice(self, output, axis, idx):
    """
    Create an expression which acts as an LHS output location 
    for a slice throught the variable 'output' along the given axis
    """
    r = self.rank(output)
    if r > 1:
      output_indices = self.build_slice_indices(r, axis, idx)
    elif r == 1:
      output_idx = self.slice_value(idx, self.none, self.int(1))
      output_indices = self.tuple([output_idx])
    else:
      output_idx = self.slice_value(self.none, self.none, self.none)
      output_indices = self.tuple([output_idx])
    return self.index(output, output_indices)
  

  def size_along_axis(self, value, axis):
    return self.shape(value, axis)

  def check_equal_sizes(self, sizes):
    pass
  
  def index(self, arr, idx, temp = False, name = None):
    """Index into array or tuple differently depending on the type"""
    
    temp = temp or name is not None
    
    arr_t = arr.type

    if isinstance(arr_t, ScalarT):
      # even though it's not correct externally, it's
      # often more convenient to treat indexing
      # into scalars as the identity function.
      # Just be sure to catch this as an error in
      # the user's code earlier in the pipeline.
      return arr
    
    if isinstance(arr_t, TupleT):
      if isinstance(idx, Const):
        idx = idx.value

      assert isinstance(idx, int), \
          "Index into tuple must be an integer, got %s" % idx
      if isinstance(idx, Const):
        idx = idx.value
      proj = self.tuple_proj(arr, idx)
      if temp:
        return self.assign_name(proj, "tuple_elt%d" % idx if name is None else name)
      else:
        return proj

    if self.is_tuple(idx):
      indices = self.tuple_elts(idx)
    elif isinstance(idx, (list,tuple)) or hasattr(idx, '__iter__'):
      indices = tuple(map(wrap_if_constant,idx))
    else:
      indices = (wrap_if_constant(idx),)

    n_required = arr_t.rank
    n_indices = len(indices)
    if n_indices < n_required:
      # all unspecified dimensions are considered fully sliced
      extra = (slice_none,) * (n_required - n_indices)
      indices = indices + extra

    if len(indices) > 1:
      idx = self.tuple(indices, name = name)
    else:
      idx = indices[0]

    t = arr_t.index_type(idx.type)
    idx_expr = Index(arr, idx, type=t)
    if temp:
      return self.assign_name(idx_expr, "array_elt" if name is None else name)
    else:
      return idx_expr

  def index_along_axis(self, arr, axis, idx, name=None):
    if arr.type.__class__ is not ArrayT:
      return arr
    assert isinstance(axis, int), \
        "Axis must be a known constant int, got: " + str(axis)
    indices = []
    for i in xrange(arr.type.rank):
      if i == axis:
        indices.append(wrap_if_constant(idx))
      else:
        indices.append(slice_none)

    index_tuple = self.tuple(indices)


    result_t = arr.type.index_type(index_tuple.type)
    idx_expr = Index(arr, index_tuple, type=result_t)
    if name:
      return self.assign_name(idx_expr, name)
    else:
      return idx_expr

  def setidx(self, arr, idx, v):
    self.assign(self.index(arr, idx, temp=False), v)


  def array_view(self, data, shape, strides, offset, nelts):
    assert isinstance(data.type, PtrT), \
        "Data field of array must be a pointer, got %s" % data.type
    if data.__class__ is not Var:
      data = self.assign_name(data, "data_ptr")
      
    if isinstance(shape.type, ScalarT):
      shape = self.tuple([shape])
    assert isinstance(shape.type, TupleT), \
      "Shape of array must be a tuple, got: %s" % shape.type
      
    if isinstance(strides.type, ScalarT):
      strides = self.tuple(strides)
      
    assert isinstance(strides.type, TupleT), \
      "Strides of array must be a tuple, got: %s" % strides.type

    ndims = len(strides.type.elt_types)
    assert ndims == len(shape.type.elt_types)

    elt_t = data.type.elt_type
    array_t = make_array_type(elt_t, ndims)
    return ArrayView(data = data, shape = shape, strides = strides, 
                     offset = offset, size = nelts, 
                     type = array_t)

  
    


from ..ndtypes import make_tuple_type

from expr import Const 
from seq_expr import SeqExpr

class Tuple(SeqExpr):
  def __init__(self, elts, type = None, source_info = None):
    self.elts = tuple(elts)
    self.type = type 
    self.source_info = source_info 
    
    if self.type is None and all(e.type is not None for e in self.elts):
      self.type = make_tuple_type(tuple(e.type for e in self.elts))
  
  def __iter__(self):
    return iter(self.elts)
  
  def __len__(self):
    return len(self.elts)
  
  def __getitem__(self, idx):
    return self.elts[idx]
  
  def __str__(self):
    if len(self.elts) > 0:
      return ", ".join([str(e) for e in self.elts])
    else:
      return "()"
    
  def children(self):
    return self.elts

  def __hash__(self):
    return hash(tuple(self.elts))


class TupleProj(SeqExpr):
  def __init__(self, tuple, index, type = None, source_info = None):
    self.tuple = tuple 
    self.index = index 
    self.type = type 
    self.source_info = source_info 

    if self.type is None:
      if self.tuple.type is not None and self.index.__class__ is Const:
        self.type = self.tuple.type.elt_types[self.index.value]
            
  def __str__(self):
    return "TupleProj(%s, %d)" % (self.tuple, self.index)

  def children(self):
    return (self.tuple,)

  def __hash__(self):
    return hash((self.tuple, self.index))


import config

import package_info 
__author__ = package_info.__author__
__email__ = package_info.__email__
__desc__ = package_info.__desc__
__license__ = package_info.__license__
__version__ = package_info.__version__
__website__ = package_info.__website__

from ndtypes import * 

from analysis import SyntaxVisitor, verify

from builder import Builder, build_fn, mk_identity_fn, mk_cast_fn

from transforms import clone_function, Transform


  
from lib import * 
from prims import *

from frontend import jit, macro, run_python_fn, run_untyped_fn, run_typed_fn
from frontend import typed_repr, specialize, find_broken_transform






# Original authors: Nathan Faggian, Stefan van der Walt, Aron Ahmadia, Olivier Grisel
# https://github.com/stefanv/growcut_py
# ...simplified and made more compact for Parakeet 

import numpy as np
import parakeet 

@parakeet.jit 
def growcut(image, state, window_radius):
    height = image.shape[0]
    width = image.shape[1]
    def attack_cell(i,j):
        winning_colony = state[i, j, 0]
        defense_strength = state[i, j, 1]
        for jj in xrange(max(0, j -  window_radius), 
                         min(width, j+window_radius+1)):
            for ii in xrange(max(0, i - window_radius), 
                             min(height, i + window_radius + 1)):
                if ii != i or jj != j:
                    ssd = np.sum( (image[i,j,:] - image[ii, jj, :]) ** 2)
                    gval = 1.0 - np.sqrt(ssd) / np.sqrt(3.0)
                    attack_strength = gval * state[ii, jj, 1]
    
                    if attack_strength > defense_strength:
                        defense_strength = attack_strength
                        winning_colony = state[ii, jj, 0]
        return np.array([winning_colony, defense_strength])
    return np.array([[attack_cell(i, j) 
                      for i in xrange(height)] 
                      for j in xrange(width)])

    
N = 50
dtype = np.double
image = np.zeros((N, N, 3), dtype=dtype)
state = np.zeros((N, N, 2), dtype=dtype)

# colony 1 is strength 1 at position 0,0
# colony 0 is strength 0 at all other positions
state[0, 0, 0] = 1
state[0, 0, 1] = 1

window_radius = 10

state_next = growcut(image, state, window_radius)

